/************************************************************************************/
/* Copyright ?????????                                                              */
/* All rights reserved.                                                             */  
/*                                                                                  */
/************************************************************************************/
/*                                                                                  */
/* AUTHOR: Aby Louw                                                                 */
/* Email:                                                                           */
/* MODIFIED: Feb 10, 2004                                                           */
/*                                                                                  */
/************************************************************************************/

#include <stdio.h>
#include "momel.h"

//constructors
Momel::Momel()
{
    Momel(false,5,30,20,50,500,1);
}


Momel::Momel(bool deb, float d, int aw, int rw, int f0min, int f0max, int vers)
{
    SetDebug(deb);
    SetDelta(d);
    SetAnalysisWindow(aw);
    SetPartitionWindow(rw);
    SetF0min(f0min);
    SetF0max(f0max);
    SetVersion(vers);
    mpoints = 0;
}

//destructor
Momel::~Momel()
{
    //nothing yet
}

void Momel::initialize()
{
    //     SetDelta(5);
    //     SetAnalysisWindow(30);
    //     SetPartitionWindow(20);
    //     SetF0min(50);
    //     SetF0max(500);
    //     SetVersion(1);
    mpoints = 0;
}


/* Set the debug value */
void Momel::SetDebug(bool deb)
{
    DEBUG = deb;
}
/* Set the delta value */
void Momel::SetDelta(float d)
{
    dlta = d;
}

/* Set the analysis window length */
void Momel::SetAnalysisWindow(int aw)
{
    A = aw;
}

/* Set the partitionning window length */
void Momel::SetPartitionWindow(int pw)
{
    R = pw;
}

/* Set the minimum f0 threshold */
void Momel::SetF0min(int fmin)
{
    hzmin = fmin;
}

/* Set the maximum f0 threshold */
void Momel::SetF0max(int fmax)
{
    hzmax = fmax;
}


/* Set the version of the output */
void Momel::SetVersion(int vers)
{
    if (vers == 1)
    {
	version = 1;
	interval = A;
    }
    else if (vers == 2)
    {
	version = 2;
	interval = (int)ceil(A/2.0);
    }
    /* should not get any other input because */
    /* of error checking in main routine      */
}

/* Get the f0 values from a EST track */
void Momel::GetF0fromTrackFile(EST_String filename)
{
    EST_Track f0input;

    if (f0input.load(filename) != read_ok) 
	EST_error("can't load \"%s\" track file\n",filename.str()); 
    
    if (f0input.num_frames() > NVALM)
	EST_error("Too many F0 values (max: %d)\n",NVALM); 

    if (!f0input.equal_space())
	EST_error("Track needs to have equal (ie. fixed) frame spacing\n"); 
    
    if (fabs(f0input.shift() - 0.01) > 0.0001)
    {
	EST_error("f0 sampling frequency does not equal 100 hz (10ms)\n");
    }
    else
    {
	nval = f0input.num_frames(); /* set the number of f0 values */
	f0.resize(nval);
	f0.zero();

	int channel_number = 0;

	if (f0input.has_channel("f0"))
	{
	    channel_number = f0input.channel_position("f0");
	}
	else if (f0input.has_channel("pitch"))
	{
	    channel_number = f0input.channel_position("pitch");
	}
	else if (f0input.has_channel("F0"))
	{
	    channel_number = f0input.channel_position("F0");
	}
	else
	{
	    EST_warning("Not sure which channel is f0, using channel 0\n");
	    channel_number = 0;
	}

	for (int x = 0; x < nval; x++)
	{
	    f0.a_check(x) = f0input.a(x,channel_number);
	}
    }

}


/* Get the f0 values from an ascii file  */
void Momel::GetF0fromAsciiFile(EST_String filename)
{
    EST_Track f0input;

    if (f0input.load(filename,"ascii",0.01) != read_ok)
	EST_error("can't load \"%s\" ascii file\n",filename.str()); 
    
    if (f0input.num_frames() > NVALM)
    {
	EST_error("Too many F0 values (max: %d)\n",NVALM); 
    }
    else
    {
	nval = f0input.num_frames(); /* set the number of f0 values */
	f0.resize(nval);
	f0.zero();

	for (int x = 0; x < nval; x++)
	{
	    f0.a_check(x) = f0input.a(x,0);
	}
    }
}

/* Save the final momel markers to file */
void Momel::SaveMomelMarkers(EST_String filename, EST_String type)
{
    for (int i = 0; i < mpoints; i++)
    {
	momel_points.t(i) = momel_points.t(i)/100;
    }

    EST_Track temp;
    momel_points.sub_track(temp,0,mpoints,0,1);
    
    if (temp.save(filename,type) != write_ok)
	EST_error("could not save ouput file \"%s\"\n",filename.str());
}


/* Save the final momel quadratic spline curve to file */
void Momel::SaveMomelSpline(EST_String filename, EST_String type)
{
    momel_curve.set_equal_space(true);

    if (momel_curve.save(filename,type) != write_ok)
	EST_error("could not save momel spline file \"%s\"\n",filename.str()); 
}


/* Print the input variables */
void Momel::PrintVars()
{
    cout <<"Momel Variables :"<<endl;
    cout <<"     delta       = "<<dlta<<endl;
    cout <<"     A           = "<<A<<endl;
    cout <<"     R           = "<<R<<endl;
    cout <<"     hzmin       = "<<hzmin<<endl;
    cout <<"     hzmax       = "<<hzmax<<endl;
    if (version == 1)
	cout <<"     version     = momel"<<endl;
    else
	cout <<"     version     = new"<<endl;
    cout <<endl;
}

/* Print the input f0 */
void Momel::PrintF0input()
{
    cout <<"Momel f0 input :"<<endl;
    cout <<"number of f0 entries = "<<nval<<endl;
    for (int i = 0; i < nval; i++)
    {
	cout <<i<<'\t'<<f0.a_check(i)<<endl;
    }
    cout <<endl;
}

/****************************************************/
/*          Momel Algorithm: Pre-processing         */
/****************************************************/
void Momel::pre_process()
{
    processedf0 = f0;

    int h = 0;
    int j = 0;
    float filter = 1 + dlta/100;

    for (int i = 0; i < processedf0.length(); i++)
    {
	h = i - 1;
	j = i + 1;

	if (processedf0.a_check(i) < hzmin)
	{
	    processedf0.a_check(i) = 0;
	}
	else if (((processedf0.a_check(i)/(processedf0.a_check(h)+1)) > filter) &&
		 ((processedf0.a_check(i)/(processedf0.a_check(j)+1)) > filter))

	{
	    processedf0.a_check(i) = 0;
	}
    }
}

/******************************************************/
/*     Momel Algorithm: Target candidate estimation   */
/******************************************************/
void Momel::get_target_candidates()
{
    int halfw = (int)ceil(A/2.0); /* half window */
    float filter = 1 + dlta/100;
  

    /* Initializing polynomial values */
    float a0 = 0;
    float a1 = 0;
    float a2 = 0;

    /* Initializing the momel time and f0 vectors */
    momel_time.resize(nval);
    momel_time.zero();
    
    momel_f0.resize(nval);
    momel_f0.zero();
 
    /* Create the legf0 vector, contains a marker to legitimate f0 values */
    EST_FVector legf0;
    legf0.resize(nval);
    legf0.zero();

    /* quadratic regression vector */
    EST_FVector qaudres;
    qaudres.resize(nval);
    qaudres.zero();

    if (DEBUG)
    {
	cout <<"processed f0: "<<endl;
    }

    for (int i = 0; i < nval; i++)
    {
	/* voicing threshold */
	if (processedf0.a_check(i) > VOICE)
	    legf0.a_check(i) = 1;

	if (DEBUG)
	{
	    cout <<i<<'\t'<<processedf0.a_check(i)<<'\t'<<legf0.a_check(i)<<endl;
	}
	/* no need to set the 0 values, as legf0 vector was initialised with 0's */
    }

    /* some temporary windows to other vectors */
    VectorWindow winleg;                /* window of the legf0 vector */
    VectorWindow winf0;                 /* window of the f0 vector */
    VectorWindow winquad;               /* window of the quadratic regression vector */


    for (int i = 0; i < nval; i++)
    {
	/* START and END windows */
	int sx = i - halfw;
	int ex = sx + A;
	int nsup = 0;
	int nsupr = -1;
	
	/* some temporary windows to other vectors */
	VectorWindow winleg(legf0,i,A);               /* window of the legf0 vector */
	VectorWindow winf0(processedf0,i,A);          /* window of the f0 vector */
	VectorWindow winquad(qaudres,i,A);            /* window of the quadratic regression vector */

	int error = 0;
	
	while ((nsup > nsupr) && (error == 0))
	{
	    nsupr = nsup;
	    nsup = 0;

	    /* calculate quadratic regression candidates */
	    error = calc_quad_regres(winleg,winf0,winquad,sx,ex,&a0,&a1,&a2);
	    
	    if (error == 1)
	    {
		/* no quadratic regression candidates */
		break;  /* break while loop */
	    }
 	    
	    for (int k = sx; k <= ex; k++)
	    {
		/* eliminate all values of f0 that differ by       */
		/* more than a delta value the f0 value estimated  */
		/* by the quadratic regression                     */

		if ((winf0.getvalue(k) == 0) ||
		    ((winquad.getvalue(k)/winf0.getvalue(k)) > filter))
		{
		    winleg.setvalue(k,0);
		    nsup += 1;
		}
	    }
	}
	winquad.win2vector(qaudres);

	/* now get target points, <t,h>, where t */
	/* is new time and h is new frequency    */
	/*           t = -a1/2*a2                */
	/*           h = a0 + a1*t + a2*t^2      */

	float tnew = 0;
	float hnew = 0;

	if ((error == 0) && (a2 != 0))
	{
	    /* everything is alright */
	    float t_temp = 0;
	    float h_temp = 0;

	    t_temp = -a1/(2 * a2);

	    if ((t_temp > (i - interval)) && (t_temp < (i + interval)))
	    {
		h_temp = a0 + a1*t_temp + a2*t_temp*t_temp;

		if ((h_temp > hzmin) && (h_temp < hzmax))
		{
		    hnew = h_temp;
		    tnew = t_temp;
		}
	    }
	}
	
	momel_time.a_check(i) = tnew;   
	momel_f0.a_check(i) = hnew;   
	
	if (DEBUG)
	{
	    cout <<i<<"\terror "<<error<<" a0 = "<<a0<<" a1 = "<<a1<<" a2 = "<<a2<<endl;
	}

    }

    if (DEBUG)
    {
	cout <<"momel_(t/f0)"<<endl<<endl;
	for (int x = 0; x < nval; x++)
	{
	    cout <<x<<'\t'<<"x\t"<<momel_time.a_check(x)<<"\ty\t"<<momel_f0.a_check(x)<<endl;
	}
    }
}		  

/*************************************************************************/
/*         Momel Algorithm:  calculate quad regression routine           */
/*************************************************************************/
int Momel::calc_quad_regres(VectorWindow winleg, VectorWindow winf0, VectorWindow &winquad, int swx, int ewx, float *pa0, float *pa1, float *pa2)
{
    /* in solving the polynomial values of the */
    /* quadratic spline function we just have  */
    /* to solve the following system of linear */
    /* equations:                              */
    /*                                         */
    /* Sy = a0*n + a1*Sx + a2*Sx^2             */
    /* Sxy = a0*Sx + a1*Sx^2 + a2*Sx^3         */
    /* Sx^2y = a0*Sx^2 + a1*Sx^3 + a2*Sx^2     */

    /* intialize variables */
    int num_voiced = 0;
    int voiced = 0;
    int error = 0;
   
    /* regression variables */
    float y = 0;         /* f0 value     */
    double x = 0;        /* x value      */
    double x_2 = 0;      /* x^2, squared */
    double x_3 = 0;      /* x^3          */
    double x_4 = 0;      /* x^4          */
    double xy = 0;       /* x*y          */ 
    double x_2y = 0;     /* x^2*y        */

    /* summation variables */
    double sum_x = 0;    /* sum of x values (time)      */
    double sum_x_2 = 0;   /* sum of x^2 (squared) values */
    double sum_x_3 = 0;   /* sum of x^3 values           */
    double sum_x_4 = 0;   /* sum of x^4 values           */
    double sum_y = 0;    /* sum of y (f0) values        */ 
    double sum_xy = 0;   /* sum of x*y values           */ 
    double sum_x_2y = 0;  /* sum of x^2*y values         */

    
    for (int i = swx; i <= ewx; i++)
    {
	voiced = (int)winleg.getvalue(i);

	if (voiced != 0)
	{
	    y = winf0.getvalue(i);
	    x = (double)i;
	    x_2 = x * x;
	    x_3 = x_2 * x;
	    x_4 = x_2 * x_2;
	    xy = x * y;
	    x_2y = x_2 * y;
	    
	    /* the sumations */
	    num_voiced += voiced;   /* number of valid f0 values */
    
	    sum_x += x;
	    sum_x_2 += x_2;
	    sum_x_3 += x_3;
	    sum_x_4 += x_4;
	    sum_y += y;
	    sum_xy += xy;
	    sum_x_2y += x_2y;
	}
    }

    if (num_voiced < 3)
    {  
	/* this eliminates erratic values after unvoiced regions */
	error = 1;
	return error;
    }
        
    /* temp variables used for calculating polynomial coefficients */
    double t8 = sum_y/sum_x_2;
    double t7 = sum_x/sum_x_2;
    double t6 = num_voiced/sum_x_2;
    double t5 = t7 - sum_x_3/sum_x_4;
    double t4 = (sum_x/sum_x_3 - t6)/(t7 -  sum_x_2/sum_x_3);
    double t3 = (t8 - sum_xy/sum_x_3)/(t7 - sum_x_2/sum_x_3);
    double t2 = (sum_x_2/sum_x_4 - t6)/t5 - t4;
    double t1 = (t8 - sum_x_2y/sum_x_4)/t5 - t3;

    (*pa0) = 0;
    (*pa1) = 0;
    (*pa2) = 0;

    (*pa0) = -t1/t2;
    (*pa1) = t3 + t4*(*pa0);
    (*pa2) = t8 - t6*(*pa0) - t7*(*pa1);
    
 
    for (int i = swx; i <= ewx; i++)
    {
	float temp = (*pa0) + (*pa1)*(float)i + (*pa2)*(float)i*(float)i;
	winquad.setvalue(i,temp);
    }
    
    return error;
}


/*************************************************************************/
/*          Momel Algorithm:  partitionning routine                      */
/*************************************************************************/
void Momel::partition()
{
    int halfw = (int)ceil(R/2.0);  /* half window */
  
    /* h and t distance vectors */
    EST_FVector hd;
    hd.resize(nval);
    hd.fill(-1);
    
    EST_FVector td;
    td.resize(nval);
    td.fill(-1);

    /* initialize the distance vector */
    EST_FVector dist;
    dist.resize(nval);
    dist.fill(-1);

    int left_start = 0;   /* starting point of left hand side window (end point is counter i)  */
    int right_end = 0;    /* end point of right hand side window (starting point is counter i) */
   
    /* sum variables for use in calculation of means */
    float sum_td = 0;
    float sum_hd = 0;
    int num_thd = 0;
   

    for (int i = 0; i < nval; i++)
    {
	left_start = 0;
	
	if (i > halfw)
	    left_start = i - halfw;
	
	right_end = nval-1;
	
	if ((i + halfw) < nval)
	    right_end = i + halfw;

	/* left side */
	float sum_t_left = 0;
	float sum_h_left = 0;
	int num_left = 0;

	for (int j = left_start; j <= i; j++)			   
	{
	    if (momel_f0.a_check(j) > hzmin)
	    {
		sum_t_left += momel_time.a_check(j);
		sum_h_left += momel_f0.a_check(j);
		num_left++;
	    }
	}

	/* right side */
	float sum_t_right = 0;
	float sum_h_right = 0;
	int num_right = 0;

	for (int j = i + 1; j <= right_end; j++)
	{
	    if (momel_f0.a_check(j) > hzmin)
	    {
		sum_t_right += momel_time.a_check(j);
		sum_h_right += momel_f0.a_check(j);
		num_right++;
	    }
	}


	/* If cible on both left and right sides */
	if ((num_left * num_right) > 0)
	{
	    td.a_check(i) = fabs(sum_t_left/num_left - sum_t_right/num_right);
	    hd.a_check(i) = fabs(sum_h_left/num_left - sum_h_right/num_right);
	    sum_td += td.a_check(i);
	    sum_hd += hd.a_check(i);
	    num_thd++;
	}
    }							   /* for loop i */

    /* get 1/mean of td and hd */
    float invmean_td = (num_thd*1.0)/sum_td;
    float invmean_hd = (num_thd*1.0)/sum_hd;
 

    /* calculate combined distance */
    for (int i = 0; i < nval; i++)
    {
	if (td.a_check(i) > 0)
	{
	    dist.a_check(i) = (td.a_check(i)*invmean_td + hd.a_check(i)*invmean_hd);
	}
    }

    if (DEBUG)
    {
	cout <<"distance vector"<<endl;
	for (int mmm = 0; mmm <  dist.length(); mmm++)
	{
	    cout <<mmm<<'\t'<<dist.a_check(mmm)<<endl;
	}
    }

    /* get boundary of partition, respecting the three conditions: */
    /*        (1)      d(x) > d(x - 1)                             */
    /*        (2)      d(x) > d(x + 1)                             */
    /*        (3)      d(x) > mean(d(x))                           */


  
    int num_part = 0;                 /* number of partitions */
    int num_dist = 0;                 /* number of legitimate distance points */
    float mean_dist = 0;              /* mean of legitimate points in distance vector */
    

    /* get the mean of the distance fector */
    for (int i=0; i < dist.length(); i++)
    {
	if (dist.a_check(i) > 0)
	{
	    mean_dist += dist.a_check(i);
	    num_dist++;
	}
    }
    mean_dist = mean_dist/num_dist;
    mean_dist = 2; //temp hack

    /* initialize partition vector */
    partitions.resize(num_part + 1);
    partitions.a_check(num_part) = 0;

    int localmax = 0;
    bool upword = false;

    /* condition (3) */
    /* search for peaks in distance vector */
    for (int i = 1; i < nval; i++)
    {
	if (!upword)
	{
	    if (dist.a_check(i) > mean_dist)          /* condition (2) */
	    {
		/* curve is upword bound */
		upword = true;
		localmax = i;
	    }
	}
	if (upword)
	{
	    if (dist.a_check(i) > dist.a_check(localmax))
	    {
		/* still upword bound */
		localmax = i;
	    }
	    if (dist.a_check(i) < mean_dist)           /* condition (3) */
	    {
		num_part++;
		partitions.resize(num_part + 1);
		partitions.a_check(num_part) = localmax;
		upword = false;
	    }
	}
    }


    /* make last dist value a partition */
    if (upword)
    {
	/* was still upword bound */
	num_part++;
	partitions.resize(num_part + 1);
	partitions.a_check(num_part) = localmax;
    }

    if (localmax < nval)
    {
	num_part++;
	partitions.resize(num_part + 1);
	partitions.a_check(num_part) = nval - 1;
    }
    
      
    if (DEBUG)
    {
	cout <<num_part<<" partitions"<<endl;
	
	for (int k = 0; k < partitions.length() ; k++)
	{
	    cout <<k<<'\t'<<partitions.a_check(k)<<endl;
	}
    }
}

/*************************************************************************/
/*          Momel Algorithm:  reduction routine                          */
/*************************************************************************/
void Momel::reduce()
{
    /* get boundary of partition, respecting the three conditions: */
    /*        (1)      d(x) > d(x - 1)                             */
    /*        (2)      d(x) > d(x + 1)                             */
    /*        (3)      d(x) > mean(d(x))                           */

  
 
    
    /* summation variables */
    double sum_x = 0;     /* sum of x values             */
    double sum_x_2 = 0;   /* sum of x^2 (squared) values */
    double sum_y = 0;     /* sum of y values             */ 
    double sum_y_2 = 0;   /* sum of y^2 (squared) values */
    
    
    int n = 0;            /* counter                     */
    double var_x = 0;     /* variance x                  */
    double var_y = 0;     /* variance y                  */
    double mean_x = 0;    /* variance x                  */
    double mean_y = 0;    /* variance y                  */
    
    int num_part = partitions.length();
    int counter = 0;

    /* current partition */
    int parstart = 0;
    int parstop = 0;

    /* starting number of momel points */
    mpoints = -1;                   
    
    for (int i = 0; i < (num_part-1); i++)
    {
	/* Current partition */
	parstart = partitions.a_check(i);		  
	parstop = partitions.a_check(i + 1);
	
	counter = 0;

	while (counter < 2) 
	{
	    counter++;

	    /* calculate means */
	    sum_x = 0;     
	    sum_x_2 = 0;   
	    sum_y = 0;     
	    sum_y_2 = 0; 
	    n = 0;
	    
	    for (int j = parstart; j <= parstop; j++)		
	    {
		if (momel_f0.a_check(j) > 0)
		{
		    sum_x += momel_time.a_check(j);            
		    sum_x_2 += momel_time.a_check(j) * momel_time.a_check(j);
		    sum_y += momel_f0.a_check(j);
		    sum_y_2 +=  momel_f0.a_check(j) *  momel_f0.a_check(j);
		    n++;
		}
	    }						   

	    if (n > 1) 
	    {

		/* mean */
		mean_x = (double)sum_x/(double)n;
		mean_y = (double)sum_y/(double)n;

		/* variance */
		var_x = (sum_x_2/n) - (mean_x*mean_x);
		var_y = (sum_y_2/n) - (mean_y*mean_y);

		if (var_x <= 0)	
		{
		    /*  case where variance should be +epsilon */
		    var_x = 0.1;
		}

		if (var_y <= 0)
		{
		    /*  case where variance should be +epsilon */
		    var_y = 0.1;
		}

		float std_x = sqrt(var_x);           /* one standard deviation x */
		float std_y = sqrt(var_y);           /* one standard deviation y */
		float start_std_x = mean_x - std_x;
		float end_std_x = mean_x + std_x;
		float start_std_y = mean_y - std_y;
		float end_std_y = mean_y + std_y;

		for (int j = parstart; j <= parstop; j++)	
		{
		    /* elimination */
		    if ((momel_f0.a_check(j) > 0) &&
			((momel_time.a_check(j) < start_std_x) || (momel_time.a_check(j) > end_std_x)
			 || (momel_f0.a_check(j) < start_std_y) || (momel_f0.a_check(j) > end_std_y)))
		    {
			momel_time.a_check(j) = 0;
			momel_f0.a_check(j) = 0;
		    }
		}					 
	    
	    }	
		 
	}
    
	
	/* recalculate means */
	sum_x = 0;
	sum_y = 0;
	n = 0;
  
	/* temporary momel points */
	float point_x = 0;
	float point_y = 0;
	float point_p = 0;

	for (int j = parstart; j <= parstop; j++)
	{
	    if (momel_f0.a_check(j) > 0)
	    {
		sum_x += momel_time.a_check(j);                 
		sum_y += momel_f0.a_check(j);                    
		n++;
	    }
	}

	if (n > 0)
	{
	    point_x = sum_x/n;
	    point_y = sum_y/n;
	    point_p = n;
	    
	    if (DEBUG)
		cout <<"mpoints "<<mpoints<<" x "<<point_x<<" y "<<point_y<<
		    " p "<<point_p<<endl<<endl;
	    
	    if (mpoints < 0)
	    {
		mpoints++;
		momel_points.resize(mpoints+1,2);
		momel_points.t(mpoints) = point_x;
		momel_points.a(mpoints,0) = point_y;
		momel_points.a(mpoints,1) = point_p;
	    }
	    else
	    {
		if (point_x > momel_points.t(mpoints))
		{
		    /*  If momel_points().x are not strictly bigger than the previous */
		    /*  the less weighted value is deleted                            */
		    mpoints++;
		    momel_points.resize(mpoints+1,2);
		    momel_points.t(mpoints) = point_x;
		    momel_points.a(mpoints,0) = point_y;
		    momel_points.a(mpoints,1) = point_p;
		}
		else		       /* t <= previous */
		{
		    if (point_p > momel_points.a(mpoints,1))
		    {
			if (DEBUG)
			{
			    cout <<"eliminated: mpoint "<<mpoints
				 <<" x "<<momel_points.t(mpoints)
				 <<" y "<<momel_points.a(mpoints,0)
				 <<" p "<<momel_points.a(mpoints,1)<<endl;
			}
			/* if current p > then erase the previous one */
			momel_points.t(mpoints) = point_x;
			momel_points.a(mpoints,0) = point_y;
			momel_points.a(mpoints,1) = point_p;
		    }
		    else
		    {
			if (DEBUG)
			{
			    cout <<"eliminated: mpoint "<<mpoints<<" x "<<point_x
				 <<" y "<<point_y<<" p "<<point_p<<endl;
			}
		    }
		}
	    }  
	}  
    }

    mpoints++;
	
	
    if (DEBUG)
    {
	for (int i = 0; i<mpoints; i++)
	{
	    cout <<momel_points.t(i)/100.0<<'\t'<<momel_points.a(i,0)<<endl;
	}
    }
}

/*************************************************************************/
/*          Momel Algorithm:  calculate the momel curve                  */
/*************************************************************************/
void Momel::calc_momel_curve()
{
	   
    /* temporary target points */
    EST_FVector ft;
    ft.resize(mpoints);
    ft.zero();
    
    EST_FVector fh;
    fh.resize(mpoints);
    fh.zero();

    EST_FVector fk;
    fk.resize(mpoints);
    fk.zero();

    EST_FVector t;
    t.resize(mpoints);
    t.zero();

    EST_FVector h;
    h.resize(mpoints);
    h.zero();

    EST_FVector k;
    k.resize(mpoints);
    k.zero();

    
    for (int i = 0; i < mpoints; i++)
    {
	ft.a_check(i) = 10 * momel_points.t(i);
	fh.a_check(i) = momel_points.a(i,0);
    }

 
    int nvalm1 = mpoints;
 
    /* quadratic regression SPLINE */
    /* Inflection points */
    for (int i = 0; i < (nvalm1-1); i++)
    {
	fk.a_check(i) =  (ft.a_check(i) + ft.a_check(i+1))/2.0;
    }

    fk.a_check(nvalm1-1) = 10000; 
	
 	
    /* conversion ms */
    for (int i = 0; i < nvalm1; i++)
    {
	t.a_check(i) = conv_ms(ft.a_check(i));
	k.a_check(i) = conv_ms(fk.a_check(i));
	h.a_check(i) = fh.a_check(i);
    }

    quasp(k,h,t,momel_curve);
 
    if (DEBUG)
    {
	for (int i = 0; i < momel_curve.length(); i++)
	{
	    cout <<i<<'\t'<<momel_curve.a(i)<<endl;
	}
    }

}
	

/* millisecond conversion function  */
float Momel::conv_ms(float ms)
{
    int pas = 10;
    float out = 0;
    
    out = ms/pas + 0.5;
    return out;
}

/* calculate quadratic function output */
float Momel::funcout(int x, float uph, float hm, float upt, float tm, float kv)
{
    float outfunc = 0;

    outfunc = uph + ((hm-uph) * (x-upt) * (x-upt))/((kv-upt) * (tm-upt));
    return outfunc;
}

/* quadratic spline function */
void Momel::quasp(EST_FVector k, EST_FVector h, EST_FVector t, EST_Track &momel_curve)
{
    float y = 0;
    int index = -1;
    int nvalm1 = mpoints;
    
    int point_k0 = (int)(k.a_check(0) + 0.5);
    int x = 0; /* counter */

    EST_FVector temp_momel;
    temp_momel.resize((int)k.a_check(nvalm1-1));
  
    for (x = 0; x <= point_k0; x++)
    {
	y = funcout(x,h.a_check(0),h.a_check(1),t.a_check(0),t.a_check(1),k.a_check(0));
	index = index + 1;
	temp_momel.a_check(index) = y;
    }
 
    for (int i = 1; i <= (nvalm1 - 2); i++)
    {
	int point_ti = (int)t.a_check(i);
	int point_ki = (int)k.a_check(i);

	for (; x < point_ti; x++)
	{
	    int j = i - 1;
	    y = funcout(x,h.a_check(i),h.a_check(j),t.a_check(i),t.a_check(j),k.a_check(j));
	    index = index + 1;
	    temp_momel.a_check(index) = y;
	}

	for (; x < point_ki; x++)
	{
	    int j = i + 1;
	    y = funcout(x,h.a_check(i),h.a_check(j),t.a_check(i),t.a_check(j),k.a_check(i));
	    index = index + 1;
	    temp_momel.a_check(index) = y;
	}
    }

    int m = nvalm1 - 1;
    int point_km = (int)k.a_check(m);

    for (; x < point_km; x++)
    {
	int j = m - 1;
	y = funcout(x,h.a_check(m),h.a_check(j),t.a_check(m),t.a_check(j),k.a_check(j));
	index = index + 1;
	temp_momel.a_check(index) = y;
    }

    momel_curve.resize(nval,1);
    momel_curve.fill_time(0.01);

    for (int i = 0; i < nval; i++)
    {
	momel_curve.a(i,0) = temp_momel.a_check(i);
    }
}
