/************************************************************************************/
/* Copyright ?????????                                                              */
/* All rights reserved.                                                             */  
/*                                                                                  */
/************************************************************************************/
/*                                                                                  */
/* AUTHOR: Aby Louw                                                                 */
/* Email:                                                                           */
/* MODIFIED: Feb 10, 2004                                                           */
/*                                                                                  */
/************************************************************************************/
#ifndef __MOMEL_H__
#define __MOMEL_H__


/* EST Speech Tools includes */
#include "EST.h"
#include "EST_types.h"
#include "EST_error.h"

/* other includes */
#include "vectorwindow.h"

/* defines */
#define NVALM 100000             /* upper limit of number of f0 values */
#define VOICE 50                 /* voicing threshold                  */  

class Momel
{
  private:

    /* variables */
    /*************/
    
    float dlta;               /* percentage of accepted error */
    int A;                    /* window size for generating targets [ms] */
    int R;                    /* window size for discriminating targets [ms] */
    int hzmin;                /* minimum accepted f0 frequency [Hz] */
    int hzmax;                /* maximum accepted f0 frequency [Hz] */
    int version;              /* momel [1] / new [2] */
    int nval;                 /* number of f0 values */
    int interval;             /* window */
    int mpoints;              /* number of momel points */
    EST_FVector f0;           /* vector containing f0 values sampled at 100 Hz (10ms) */
    EST_FVector processedf0;  /* exactly what it says, processed f0 */
    EST_FVector momel_time;   /* final output, momel time points */
    EST_FVector momel_f0;     /* final output, momel f0 points */
    EST_Track   momel_points; /* final momel markers */
    EST_Track   momel_curve;  /* the momel f0 macroprosodic curve */
    EST_IVector partitions;   /* the partitioning points */
    bool DEBUG;               /* display debug info */ 
     
    /* functions */
    /*************/

    /* millisecond conversion function  */
    float conv_ms(float ms);  /* millisecond conversion function  */

    /* quadratic spline function */
    void quasp(EST_FVector k, EST_FVector h, EST_FVector t, EST_Track &momel_curve);

    /* calculate quadratic function output */
    float funcout(int x, float uph, float hm, float upt, float tm, float kv);

    /* inititialize momel variables */
    void initialize();


    
  public:
    //defualt constructor
    Momel();

    /* Initialized constructor */
    Momel(bool deb = false, float d = 5, int aw = 30, int rw = 20, int f0min = 50, int f0max = 500, int vers = 1);
    
    // Desructor
    ~Momel();

    /* Set the debug value */
    void SetDebug(bool deb);

    /* Set the delta value */
    void SetDelta(float d);

    /* Set the analysis window length */
    void SetAnalysisWindow(int aw);

    /* Set the partitionning window length */
    void SetPartitionWindow(int pw);

    /* Set the minimum f0 threshold */
    void SetF0min(int fmin);

    /* Set the maximum f0 threshold */
    void SetF0max(int fmax);

    /* Set the version of the output */
    void SetVersion(int ver);

    /* Print the input variables */
    void PrintVars();

    /* Print the input f0 */
    void PrintF0input();

    /* Get the f0 values from an EST track file */
    void GetF0fromTrackFile(EST_String filename);

    /* Get the f0 values from an ascii file */
    void GetF0fromAsciiFile(EST_String filename);

    /* Save the final momel markers to file */
    void SaveMomelMarkers(EST_String filename, EST_String type);

    /* Save the final momel quadratic spline curve to file */
    void SaveMomelSpline(EST_String filename, EST_String type);


    /************************************************************************************************************/
    /*              Start of actual Momel Algorithm as described in                                             */   
    /*    @misc{hirst93automatic,                                                                               */
    /*          author = "D. Hirst and R. Espesser",                                                            */
    /*          title  = "Automatic Modelling of Fundamental Frequency using a quadratic spline function",      */
    /*          text   = "Hirst, D., Espesser, R. (1993). Automatic Modelling of Fundamental Frequency          */
    /*                    using a quadratic spline function, Travaux de l'Institut de Phontique d'Aixen        */
    /*                    -Provence, 15, 75-85.",                                                               */
    /*          year   = "1993",                                                                                */
    /*          url    = "citeseer.nj.nec.com/hirst93automatic.html" }                                          */
    /************************************************************************************************************/

     /*	pre-processing */
    void pre_process();

    /* target candidate estimation */
    void get_target_candidates();

    /* calculate quadratic regression, return values   0 : OK (default)   1 : Problem  */
    int calc_quad_regres(VectorWindow winleg, VectorWindow winf0, VectorWindow &winquad, int swx, int ewx, float *pa0, float *pa1, float *pa2);

    /* partitionning of target candidates */
    void partition();

    /* reduction of target candidates */
    void reduce();

    /* calculate the momel curve from the momel target points */
    void calc_momel_curve();
};
    

    


#endif
